import sys
from django.conf import settings


class DjangoAutoshardSettings:
    def __init__(self):
        try:
            user_settings = settings.DJANGO_AUTOSHARD
        except AttributeError:
            user_settings = {}

        self.__settings = dict(
            NODES=dict(),
            EPOCH='2016-01-01',
            MAX_SHARDS=1 << 13,
        )
        self.__settings.update(user_settings)
        settings.DJANGO_AUTOSHARD = self

    def __getattr__(self, name):
        return self.__settings.get(name)


django_autoshard_settings = DjangoAutoshardSettings()
sys.modules[__name__] = django_autoshard_settings



from collections import OrderedDict

from django.conf import settings as django_settings
from django.core import exceptions
from .shard import Shard
from . import utils


class ShardingFactory:
    def __init__(self):
        self.primary = django_settings.DATABASES.get('default')

    def configure(self):
        shards = dict()
        nodes = django_settings.DJANGO_AUTOSHARD.NODES
        for node in nodes:
            try:
                _ = node['HOST']
            except KeyError:
                raise exceptions.ImproperlyConfigured('Node {} does not have a host.'.format(node))
            try:
                _ = node['RANGE']
            except KeyError:
                raise exceptions.ImproperlyConfigured('Node {} does not specify a shard range.'.format(node))

            node_shards = self.set_logical_shards(node)
            shards.update(node_shards)

        django_settings.SHARDS = OrderedDict(sorted(shards.items()))
        django_settings.DATABASE_ROUTERS = ('django_autoshard.routers.ShardRouter', )

    def set_logical_shards(self, node):
        result = dict()
        for i in node['RANGE']:
            shard = self.primary.copy()
            shard['HOST'] = node['HOST']
            shard['NAME'] = '{}_{}'.format(self.primary['NAME'], i)
            alias = 'shard_{}'.format(i)
            django_settings.DATABASES[alias] = shard
            node_index = utils.get_shard_index(alias)
            replicas = self.set_replicas()
            result[node_index] = Shard(node_index, alias, replicas)
        return result

    def set_replicas(self):
        return {}



from django.contrib.auth import get_user_model
from django.db import models

from django_autoshard.querysets import ShardedQuerySet


class ShardedManager(models.Manager):
    _queryset_class = ShardedQuerySet

    def all(self):
        return self.filter()


class UserManager(ShardedManager):
    def create_user(self, username, email, password=None, save=True):  # pragma: no cover
        user = get_user_model()(
            username=username,
            email=email)
        user.set_password(password)
        if save:
            user.save()
        return user

    def create_superuser(self, username, email, password):  # pragma: no cover
        user = self.create_user(username, email, password, False)
        user.is_active = True
        user.is_staff = True
        user.is_superuser = True
        user.save()
        return user




from django.conf import settings
from django.db import models
from django_autoshard import utils


class ShardedQuerySet(models.QuerySet):
    def filter(self, *args, **kwargs):
        if not hasattr(settings, 'SHARDS') or len(settings.SHARDS) == 0:
            return super(ShardedQuerySet, self).filter(*args, **kwargs)

        if len(args) == 0 and len(kwargs) == 0:
            return self.all()

        if 'pk' in kwargs:
            shard_index = utils.get_shard_index_from_uuid(kwargs['pk'])
            shard = utils.get_shard_from_index(shard_index)
            self._db = shard.alias
        elif self.model.SHARD_KEY in kwargs:
            shard = utils.get_shard(kwargs[self.model.SHARD_KEY])
            self._db = shard.alias
        return super(ShardedQuerySet, self).filter(*args, **kwargs)

    def count(self):
        if self._db is not None:
            return super().count()
        raise NotImplementedError('ShardedQuerySet does not implement count().')

    def all(self):
        if self._db is not None:
            return super().all()
        raise NotImplementedError('ShardedQuerySet does not implement all().')



from itertools import chain
import time
import six

from django.conf import settings
from django.core.management.commands.migrate import Command as DjangoMigrateCommand
from django.db import OperationalError


class Command(DjangoMigrateCommand):
    def migrate(self, shard, *args, **options):
        try:
            start = time.time()
            options['database'] = shard.alias
            super(Command, self).handle(*args, **options)
            shard.connection.close()
            self.stdout.write(
                self.style.MIGRATE_HEADING('Migrated database %s in %.4f sec' % (shard.database, time.time() - start)))
        except OperationalError as e:
            self.stdout.write(
                    self.style.MIGRATE_HEADING('Failed migrating database: %s. Error: %s' % (options['database'], e)))

    def run(self, *args, **options):
        for _, shard in settings.SHARDS.items():
            self.migrate(shard, *args, **options)

    def handle(self, *args, **options):
        options['verbosity'] = 0
        try:
            self.run(*args, **options)
            self.stdout.write(self.style.MIGRATE_HEADING('Done.\n'))
        except KeyboardInterrupt:
            self.stdout.write(self.style.MIGRATE_HEADING('Migration interrupted.\n'))



from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import ProgrammingError


class Command(BaseCommand):
    help = 'Create shards databases'

    def handle(self, *args, **opts):
        for _, shard in settings.SHARDS.items():
            cursor = shard.connection._nodb_connection.cursor()
            try:
                cursor.execute('CREATE DATABASE %s' % shard.database)
                self.stdout.write(self.style.MIGRATE_HEADING(
                    'Database [%s] created.' % shard.database
                ))
            except ProgrammingError as e:
                self.stderr.write(str(e))
        self.stdout.write(self.style.MIGRATE_HEADING('Done.\n'))



import bisect
import time
import hashlib
import random

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from . import settings
from django.utils import timezone


def get_shard_index(key):
    _hash = hashlib.md5(key.encode()).hexdigest()
    return int(_hash, 16) % settings.MAX_SHARDS


def get_shard_index_from_uuid(uuid):
    return (uuid >> 10) & 0x1FFF


def get_shard(key):
    index = get_shard_index(key)
    return get_shard_from_index(index)


def get_shard_from_index(index):
    try:
        return django_settings.SHARDS[index]
    except KeyError:
        shard_indexes = list(django_settings.SHARDS.keys())
        pos = bisect.bisect(shard_indexes, index)
        if pos >= len(shard_indexes):
            pos = 0
        index = shard_indexes[pos:pos + 1][0]
        return django_settings.SHARDS[index]


def generate_uuid(shard_index):
    try:
        epoch = timezone.datetime.strptime(settings.EPOCH, '%Y-%m-%d')
        epoch = time.mktime(epoch.timetuple()) * 1000
    except ValueError:
        raise ImproperlyConfigured('EPOCH must be in Y-m-d format')
    local_id = random.randint(0, 1204)
    now = time.time() * 1000
    result = int(now - epoch) << 23
    result |= shard_index << 10
    result |= local_id % 1024
    return result



from django.apps import apps
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.db import connections
from django.db.utils import OperationalError, ProgrammingError
from django_autoshard import models


class Command(BaseCommand):
    help = 'Create shards databases'

    def add_arguments(self, parser):
        parser.add_argument('--list', '-l', action='store_true', default=False,
                            help='Show a list of all the constraints that will be dropped')

    def handle(self, *args, **options):
        for connection in connections.all():
            with connection.cursor() as cursor:
                for model in apps.get_models():
                    if issubclass(model, models.ShardedModel):
                        continue
                    self.run(connection, cursor, model, **options)

    def run(self, connection, cursor, model, **options):
        try:
            constraints = dict()
            for key, val in cursor.db.introspection.get_constraints(cursor, model._meta.db_table).items():
                if issubclass(model, models.ShardRelatedModel):
                    if val['foreign_key'] is not None and val['foreign_key'][0] == get_user_model()._meta.db_table:
                        continue
                    elif val['foreign_key'] is not None:
                        constraints[key] = val

                else:
                    if val['foreign_key'] is not None and val['foreign_key'][0] == get_user_model()._meta.db_table:
                        constraints[key] = val
        except ProgrammingError:
            return

        if len(constraints) == 0:
            f = self.style.MIGRATE_HEADING('[{}] No constraints defined for {}.'.format(connection.alias, str(model)))
            self.stdout.write(f)
            return

        for key, val in constraints.items():
            db_table, _ = val['foreign_key']

            sql = cursor.db.SchemaEditorClass.sql_delete_fk % dict(
                table=model._meta.db_table,
                name=key
            )
            if options.get('list'):
                f = self.style.MIGRATE_HEADING('{}'.format(sql))
                self.stdout.write(f)
                continue

            try:
                f = self.style.MIGRATE_HEADING('[{}] Executing {}'.format(connection.alias, sql))
                self.stdout.write(f)

                cursor.execute(sql)
                f = self.style.MIGRATE_HEADING('[{}] Done.\n'.format(connection.alias))
                self.stdout.write(f)
            except OperationalError as e:
                f = self.style.MIGRATE_HEADING('[{}] Failed: {}.\n'.format(connection.alias, str(e)))
                self.stdout.write(f)



from django.db import connections


class Shard:
    def __init__(self, index, alias, replicas):
        self.__index = index
        self.__alias = alias
        self.__replicas = replicas

    @property
    def alias(self):
        return self.__alias

    @property
    def index(self):
        return self.__index

    @property
    def connection(self):
        return connections[self.alias]

    @property
    def database(self):
        return self.connection.settings_dict['NAME']

    @property
    def host(self):
        return self.connection.settings_dict['HOST']

    def __str__(self):
        return self.alias



from django.apps import AppConfig


class DjangoAutoShardApp(AppConfig):
    name = 'django_autoshard'
    verbose_name = 'Django Autoshard'

    def __init__(self, app_name, app_module):
        super(DjangoAutoShardApp, self).__init__(app_name, app_module)

    def ready(self):
        from .factory import ShardingFactory
        ShardingFactory().configure()
        super(DjangoAutoShardApp, self).ready()



import os
import re
from setuptools import setup, find_packages


def readme():
    with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
        return readme.read()


def version():
    pattern = re.compile(r'__version__ = \'([\d\.]+)\'')
    with open(os.path.join('django_autoshard', '__init__.py')) as f:
        data = f.read()
        return re.search(pattern, data).group(1)


setup(
    name='django-autoshard',
    version=version(),
    packages=find_packages(),
    include_package_data=True,
    license='BSD',
    description='A Django library that makes sharding easy.',
    long_description=readme(),
    url='https://github.com/cipriantarta/django-autoshard',
    author='Ciprian Tarta',
    author_email='me@cipriantarta.ro',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Framework :: Django",
        "Framework :: Django :: 1.7",
        "Framework :: Django :: 1.8",
        "Framework :: Django :: 1.9",
    ],
    keywords='django shard sharding',
    install_requires=[
        'django>=1.7',
        'six',
    ],
    test_suite='django_autoshard.tests',
)



from django.contrib.auth.models import AbstractUser

from django_autoshard.models import ShardedModel, ShardedManager


class User(ShardedModel, AbstractUser):
    SHARD_KEY = 'email'

    objects = ShardedManager()



from django_autoshard.models import ShardedModel, ShardRelatedModel


class ShardRouter:
    def db_for_write(self, model, **hints):
        if not issubclass(model, ShardedModel) and not issubclass(model, ShardRelatedModel):
            return 'default'
        instance = hints.get('instance')
        shard = getattr(instance, 'shard', None)
        if shard is not None:
            return shard.alias

    def allow_relation(self, obj1, obj2, **hints):
        return True



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-03-30 06:30
from __future__ import unicode_literals

from django import get_version
from django.db import migrations, models
import django.core.validators
import django.utils.timezone


def auth_dependency():
    v = get_version().split('.')
    if v[0] == '1':
        if v[1] == '10':
            return '0008_alter_user_username_max_length'
        elif v[1] == '7':
            return '0001_initial'
        elif v[1] == '8':
            return '0006_require_contenttypes_0002'
        elif v[1] == '9':
            return '0007_alter_validators_add_error_messages'


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', auth_dependency()),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('username', models.CharField(error_messages={'unique': 'A user with that username already exists.'}, help_text='Required. 30 characters or fewer. Letters, digits and @/./+/-/_ only.', max_length=30, unique=True, validators=[django.core.validators.RegexValidator('^[\\w.@+-]+$', 'Enter a valid username. This value may contain only letters, numbers and @/./+/-/_ characters.')], verbose_name='username')),
                ('first_name', models.CharField(blank=True, max_length=30, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=30, verbose_name='last name')),
                ('email', models.EmailField(blank=True, max_length=254, verbose_name='email address')),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('id', models.BigIntegerField(primary_key=True, serialize=False)),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
